/*-
 * Copyright (c) 2018 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <string.h>

#include <openssl/crypto.h>

#include "cache/cache.h"

#include "vcc_if.h"

#include "vfp.h"
#include "keys.h"

/* Event function */
int
VPFX(event)(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
	ASSERT_CLI();
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(priv);

	static int loaded = 0;

	switch(e) {
	case VCL_EVENT_LOAD:
		VRT_AddVFP(ctx, &vfp_encrypt);
		VRT_AddVFP(ctx, &vfp_decrypt);
		assert(loaded >= 0);
		if (loaded++ == 0)
			if (KEY_Init(ctx) != 0)
				return (-1);
		return (0);
	case VCL_EVENT_DISCARD:
		VRT_RemoveVFP(ctx, &vfp_encrypt);
		VRT_RemoveVFP(ctx, &vfp_decrypt);
		AN(loaded);
		if (--loaded == 0)
			KEY_Fini();
		return (0);
	case VCL_EVENT_WARM:
	case VCL_EVENT_COLD:
		return (0);
	default:
		WRONG("illegal event enum");
	}
	NEEDLESS(return (0));
}

#define CHECK_ID(ctx, id, len, ret)					\
	do {								\
		if ((id) == NULL) {					\
			VRT_fail((ctx), "key id is NULL");		\
			return ret;					\
		}							\
		len = strlen(id);					\
		if ((len) > 255) {					\
			VRT_fail((ctx), "key id \"%.80s...\" too long "	\
				 "(length %zu > 255)", (id), (len));	\
			return ret;					\
		}							\
	} while(0)

#define CHECK_KEY(ctx, key)						\
	do {								\
		if ((key) == NULL || (key)->blob == NULL) {		\
			VRT_fail((ctx), "key contents are empty");	\
			return;						\
		}							\
		if ((key)->len != 16) {					\
			VRT_fail((ctx), "illegal key length %zu (must be 16)", \
				 (key)->len);				\
			return;						\
		}							\
	} while(0)

VCL_VOID
vmod_add_key(VRT_CTX, VCL_STRING id, VCL_BLOB key)
{
	size_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	CHECK_ID(ctx, id, len,);
	CHECK_KEY(ctx, key);

	/* KEY_Add calls VRT_fail() on error. */
	(void)KEY_Add(ctx, (uint8_t *)id, (uint8_t)len, key->blob);
}

VCL_VOID
vmod_update_key(VRT_CTX, VCL_STRING id, VCL_BLOB key)
{
	size_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	CHECK_ID(ctx, id, len,);
	CHECK_KEY(ctx, key);

	/* KEY_Update calls VRT_fail() on error. */
	(void)KEY_Update(ctx, (uint8_t *)id, (uint8_t)len, key->blob);
}

VCL_VOID
vmod_set_key(VRT_CTX, VCL_STRING id, VCL_BLOB key)
{
	size_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_ID(ctx, id, len,);
	CHECK_KEY(ctx, key);

	/* KEY_Set calls VRT_fail() on error. */
	(void)KEY_Set(ctx, (uint8_t *)id, (uint8_t)len, key->blob);
}

VCL_VOID
vmod_delete_key(VRT_CTX, VCL_STRING id)
{
	size_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_ID(ctx, id, len,);

	/* KEY_Delete calls VRT_fail() on error. */
	(void)KEY_Delete(ctx, (uint8_t *)id, (uint8_t)len);
}

VCL_BOOL
vmod_key_exists(VRT_CTX, VCL_STRING id)
{
	size_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_ID(ctx, id, len, 0);

	return (KEY_Exists((uint8_t *)id, (uint8_t)len));
}

VCL_TIME
vmod_key_added(VRT_CTX, VCL_STRING id)
{
	size_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_ID(ctx, id, len, 0);

	return (KEY_Added(ctx, (uint8_t *)id, (uint8_t)len));
}

VCL_STRING
vmod_libcrypto_version(VRT_CTX)
{
	(void)ctx;
	VCL_STRING version = OpenSSL_version(OPENSSL_VERSION);
	if (version == NULL)
		return "unknown libcrypto version";
	return version;
}

VCL_STRING
vmod_version(VRT_CTX)
{
	(void) ctx;
	return VERSION;
}
